require("md5")
local debuglevel = 0 -- 0=no debug, 1=notice debug, 2=full verbose
local function debug(level, ...)
	if type(level) ~= "number" then
		if debuglevel >= 2 then
			print(level,...)
		end
	else
		if debuglevel >= level then
			print(...)
		end
	end
end

local config

local function syscall(command)
	debug(2, "Executing: "..command)
	return os.execute(command)
end

local function md5file(path)
	local fileHandle = io.open(path,"rb")
	if not fileHandle then return 0 end
	local md5sum=md5.sumhexa(fileHandle:read("*a"))
	fileHandle:close()
	return md5sum
end

local function dirExists(path)
	return ((syscall("dir \""..path.."\" >nul 2> nul"))==0)
end

local function surround(text)
	local maxLen=0
	for line in string.gmatch(text, "[^\n]+") do
		maxLen = math.max(maxLen, string.len(line)+2)
	end
	local output = ""
	for i=1, maxLen+2 do
		output=output.."="
	end
	for line in string.gmatch(text, "[^\n]+") do
		output = output.."\n="
		line = " "..line.." "
		local lineLen = string.len(line)
		for i=1, math.ceil((maxLen-lineLen)/2) do
			output = output.." "
		end
		output = output..line
		for i=1, math.floor((maxLen-lineLen)/2) do
			output = output.." "
		end
		output=output.."="
	end
	output = output.."\n"
	for i=1, maxLen+2 do
		output=output.."="
	end
	return output
end
--print(surround([[
-- This is a text with various line sizes
-- they should all be equalized so they look like a compact
-- box of text
-- even if they are very
-- small]]))
--do return end

-- returns a passed table in its lua code representation
local function dumpTable(t)
	local tableString = "{"
	for okey, value in pairs(t) do
		local thisString
		local key = type(okey) == "string" and "[\""..okey.."\"]" or "["..okey.."]"
		key = key:gsub("\\","\\\\")
		local valType = type(value)
		if (valType == "function") or (valType == "userdata") or (valType == "file") then
			thisString = "--"..key.." = <skipped "..valType..">,"
		elseif (valType == "table") then
			thisString = key.." = "..dumpTable(value)..","
		elseif (valType == "string") then
			thisString = key.." = \""..value:gsub("\\","\\\\").."\","
		else
			thisString = key.." = "..tostring(value)..","
		end
		local outString
		for line in string.gmatch(thisString, "[^\n]+") do
			outString = (outString and outString.."\n" or "").."  "..line
		end
		tableString = tableString.."\n"..outString
	end
	tableString = tableString.."\n}"
	return tableString
end
--[[local test={
	value1="a string",
	value2="another string",
	theanswer=42,
	inside={
		is="another table",
	},
	afunc=function() end
}
print(dumpTable(test))
do return end--]]
local function flushConfig()
	local confHandle = io.open("conf.slc", "w")
	confHandle:write(dumpTable(config))
	confHandle:close()
end
local driveLetters={"A","B","C","D","E","F","G","H","I","J","K","L","M","N","O","P","Q","R","S","T","U","V","W","X","Y","Z"}
local wowFilesRequired = {"battle.net.dll","blizzard downloader.exe","blizzard updater.exe","dbghelp.dll","divxdecoder.dll","ijl15.dll","launcher.exe","microsoft.vc80.crt.manifest","msvcr80.dll","repair.exe","unicows.dll","wow.exe","wowerror.exe","wow.mfil","wow.tfil"}

local scanFolders = function()
	local wowInstalls = {}
	print("Scanning for WoW install folders...")
	for _, letter in ipairs(driveLetters) do
		debug(1,"Scanning drive "..letter..":")
		local tmpName=os.tmpname()
		syscall("dir /B /S "..letter..":\\ 2> nul | find /i \"wow.exe\" > "..tmpName)
		local f = io.open(tmpName)
		local nextLine=f:read("*l")
		while nextLine do
			local startPos,endPos,path = nextLine:find("(.+)[Ww][Oo][Ww]%.exe")
			if startPos == 1 and endPos == string.len(nextLine) then
				debug(2,path.." found")
				table.insert(wowInstalls, path)
			end
			nextLine=f:read("*l")
		end
		f:close()
		os.remove(tmpName)
	end
	return wowInstalls
end

local isSymlinkInstall = function(path)
	local tmpName = os.tmpname()
	syscall("dir \""..path.."\" /A:D 2> nul | find \"Data\" > "..tmpName)
	local f=io.open(tmpName)
	local line = f:read("*l")
	local symTarget
	if line then
		debug(2, "Checking "..line.." for symlink references")
		symTarget = line:match("%<SYMLINKD%>%s+Data %[(.+\\)Data%]")
	end
	f:close()
	os.remove(tmpName)
	return symTarget
end

local syncFolders = function(path, from)
	local syncfiles = {}
	for _, file in ipairs(wowFilesRequired) do
		local srcmd, dstmd = md5file(from..file), md5file(path..file)
		debug(2, "File:",file,"Source MD5:",srcmd,"Destination MD5:",dstmd)
		if (srcmd ~= 0) and (srcmd ~= dstmd) then
			table.insert(syncfiles, file)
		end
	end
	if #syncfiles > 0 then
		print("I've found the following files to be desynchronized:")
		print(table.concat(syncfiles,", "))
		print("Synchronize those files? (y/n)")
		local response = io.read():lower()
		if response == "y" then
			for _, file in ipairs(syncfiles) do
				debug(1,"Copying: "..file)
				syscall("copy /Y \""..from..file.."\" \""..path..file.."\" > nul 2> nul")
				print("Copied: "..file)
			end
		end
	else
		print("All files are already synced")
	end
end

local function first_setup()
	print([[
=================================
===== WORLD OF WARCRAFT SYM =====
=================================
This appears to be your first
time running WoW SYM. Do you want
me to scan for any WoW installs
I can find? (y/n)]])
	local response = io.read()
	if response:lower() ~= "y" then return end
	local wowFolders = scanFolders()
	print("I have been able to find "..#wowFolders.."\nWoW installations on your machine!")
	config={managedInstalls = {nonsym={},sym={}}}
	for _, path in ipairs(wowFolders) do
		print(surround("Configuring WoW install in:\n"..path))
		local symTarget = isSymlinkInstall(path)
		if symTarget then
			print("I have detected that this installation is already SYM linked\nto an installation in "..symTarget..".\nDo you want me to synchronize the folders and add this installation\n to my known folder list? (y/n)")
			local response = io.read():lower()
			if response == "y" then
				print("Syncing files...")
				syncFolders(path, symTarget)
				config.managedInstalls.sym[path]=symTarget
				flushConfig()
				print(path.." => "..symTarget.." was synced and added to the list of known SYMlinks")
			end
		else
			print("This is a non-symlinked installation of WoW.\nDo you want me to add this installation to my list of\nWoW installations you can use to create new SYMlink installs? (y/n)")
			local response = io.read():lower()
			if response == "y" then
				table.insert(config.managedInstalls.nonsym, path)
				print(path.." was registered as a WoW installation")
				flushConfig()
			end
		end
	end
end
local configHandle = io.open("conf.slc", "r")
if (not configHandle) or (not configHandle:read(1)) then
	local confBackHandle = io.open("conf.slc.bak", "r")
	if confBackHandle and confBackHandle:read(1) then
		confBackHandle:close()
		syscall("copy conf.slc.bak conf.slc > nul 2> nul")
		configHandle = io.open("conf.slc","r")
	else
		if not first_setup() then
			return
		end
	end
end
configHandle:seek("set",0)
local stringLoad="return "..configHandle:read("*a")
confFunc,confError = loadstring(stringLoad)
if not confFunc then
	print(stringLoad)
	error(confError)
end
config = confFunc()
configHandle:close()

syscall("copy conf.slc conf.slc.bak > nul 2> nul")
local input
while input ~= "x" do
	print([[
=================================
===== WORLD OF WARCRAFT SYM =====
=================================
Welcome! Please choose an action:
R: Re-scan for WoW folders
S: Sync all symlinked WoW folders
C: Create a new WoW folder
D: Delete a symlinked WoW folder
X: Exit this program]])
	input = io.read("*l"):lower()
	if input == "s" then
		for dst, src in pairs(config.managedInstalls.sym) do
			if not dirExists(dst) then
				print("Destination directory no longer exists:\n"..dst.."\nRemoving...")
				config.managedInstalls.sym[dst] = nil
				print("Removed.")
			else
				print(surround("Syncing:\n"..dst.."\nParent:\n"..src))
				syncFolders(dst, src)
			end
			syscall("pause")
		end
		flushConfig()
	end
	if input == "c" then
		print(surround("Create sync folder\nSource folder"))
		for n, path in ipairs(config.managedInstalls.nonsym) do
			print(n..": "..path)
		end
		local selection=io.read("*n")
		local srcPath = config.managedInstalls.nonsym[selection]
		while not srcPath do
			selection=io.read("*n")
			srcPath=config.managedInstalls.nonsym[selection]
		end
		print(surround("Create sync folder\nDestination path"))
		print("Enter the location you want to create a World of Warcraft copy in:")
		local dstPath = io.read("*l")
		local cur=dstPath:match(".:")
		while not cur do
			dstPath=io.read("*l")
			cur=dstPath:match(".:")
		end
		while true do
			local new = dstPath:match(cur.."(\\[^\\]+)")
			if not new then break end
			cur=cur..new
			if not dirExists(cur) then
				syscall("mkdir "..cur.." > nul 2> nul")
				debug(2,"Directory "..cur.." does not exist")
			else
				debug(2, "Directory "..cur.." exists")
			end
		end
		cur=cur.."\\"
		print(surround("Create sync folder:\n"..cur.."\nFrom:\n"..srcPath))
		print("Okay? (y/n)")
		local response = io.read():lower()
		if response == "y" then
			print("Copying required files...")
			for _, file in ipairs(wowFilesRequired) do
				syscall("copy /Y \""..srcPath..file.."\" \""..cur..file.."\" > nul 2> nul")
				print(file.." copied")
			end
			syscall("mklink /D \""..cur.."\\Data\" \""..srcPath.."\\Data\" > nul 2> nul")
			print("SYMlink created")
			config.managedInstalls.sym[cur] = srcPath
			print("Registered with WoW SYM.")
			flushConfig()
			print("Complete.")
			syscall("pause")
		end
	end
end
flushConfig()
print("Have a nice day!")